package be.cylab.java.wowa.training;

import java.util.Arrays;
import java.util.List;
import java.util.Random;

/**
 * Abstract class Solution.
 */
public abstract class AbstractSolution
        implements Comparable<AbstractSolution>, Cloneable {

    protected double[] weights_w;
    protected double[] weights_p;
    protected double distance = Double.POSITIVE_INFINITY;

    /**
     * SolutionDistance constructor. Needs weight number as parameter
     *
     * @param weights_number
     */
    public AbstractSolution(final int weights_number) {
        this.weights_w = new double[weights_number];
        this.weights_p = new double[weights_number];
        for (int i = 0; i < weights_number; i++) {
            this.weights_w[i] = Math.random();
            this.weights_p[i] = Math.random();
        }
        this.normalize();
    }

    /**
     * Constructor solution distance. Made for test to initialize the Random
     * Generator
     *
     * @param weights_number int
     * @param seed           int
     */

    public AbstractSolution(final int weights_number, final int seed) {
        Random rnd = new Random(seed);
        this.weights_w = new double[weights_number];
        this.weights_p = new double[weights_number];
        for (int i = 0; i < weights_number; i++) {
            this.weights_w[i] = rnd.nextDouble();
            this.weights_p[i] = rnd.nextDouble();
        }
        this.normalize();
    }

    abstract void computeScoreTo(List<double[]> data, double[] expected);

    /**
     * @return
     */
    @Override
    public final String toString() {
        return "SolutionDistance{"
                + "weights_w=" + Arrays.toString(weights_w)
                + ", weights_p=" + Arrays.toString(weights_p)
                + ", distance=" + Math.abs(distance)
                + '}';
    }

    /**
     * Change a random gene in a weight vector.
     * The gene is changed is the probability is higher than a random double
     *
     * @param probability
     */
    final void randomlyMutateWithProbability(final double probability) {
        double tos = Math.random();
        if (tos > probability) {
            // do nothing
            return;
        }

        //new weight value and position
        double new_weight_value = Math.random();
        int random_index = Utils.randomInteger(0, this.weights_p.length - 1);
        //Select w or p weights
        int weight_selection = Utils.randomInteger(0, 1);
        if (weight_selection == 0) {
            this.weights_w[random_index] = new_weight_value;
        } else {
            this.weights_p[random_index] = new_weight_value;
        }
        this.normalize();

    }

    /**
     * @param solution
     * @return int
     */
    @Override
    public final int compareTo(final AbstractSolution solution) {
        if (this.getDistance() > solution.getDistance()) {
            return 1;
        } else if (this.getDistance() < solution.getDistance()) {
            return -1;
        } else {
            return 0;
        }
    }

    /**
     * Function to normalize SolutionDistance weights.
     * Weights must be between 0 and 1 and the sum of the weight in a vector
     * must be equal to 1
     */
    final void normalize() {
        this.weights_w = Utils.normalizeWeights(this.weights_w);
        this.weights_p = Utils.normalizeWeights(this.weights_p);
    }

    /**
     * @return
     */
    public final double getDistance() {
        return distance;
    }

    /**
     * @param distance
     */
    public final void setDistance(final double distance) {
        return;
    }

    /**
     * @return
     */
    public final double[] getWeightsW() {
        return this.weights_w;
    }

    /**
     * @return
     */
    public final double[] getWeightsP() {
        return this.weights_p;
    }

    /**
     * @param weights_w
     */
    final void setWeightsW(final double[] weights_w) {
        this.weights_w = weights_w;
    }

    /**
     * @param weights_p
     */
    final void setWeightsP(final double[] weights_p) {
        this.weights_p = weights_p;
    }


    /**
     * Copy.
     *
     * @return
     * @throws java.lang.CloneNotSupportedException if clone is not supported
     */
    public final SolutionDistance clone() throws CloneNotSupportedException {
        return (SolutionDistance) super.clone();
    }

}
